import click
import json
import typing as t

from tqdm import tqdm
from pathlib import Path
from concurrent import futures

from . import cli, load_scans_from_file
from .. import msg, DependencyScan

from ..api import TrustSourceAPI


@cli.command('check', help='Checks the scan result')
@cli.inout_default_options(_in=True, _out=True, _fmt=True)
@cli.api_default_options(is_project_name_required=False, project_name_help_hint="required w/o '--vulns-only'")
@click.option('--vulns-only', default=False, is_flag=True,
              help="Only check component vulnerabilities")
@click.option('--exit-on-legal/--no-exit-on-legal', default=True,
              help="Exit with code (1) in case of legal issues")
@click.option('--exit-on-vulns/--no-exit-on-vulns', default=True,
              help="Exit with code (1) in case of vulnerabilities")
@click.option('--Werror', default=False, is_flag=True,
              help="Treat warnings as violations")
@click.option('--vulns-confidence', type=click.Choice(choices=['high', 'medium', 'low']), default='medium',
              help="Confidence level while checking component vulnerabilities")
def check_scan(path: Path, output_path: t.Optional[Path], scan_format: str,
               project_name: str, base_url: str, api_key: str,
               vulns_only: bool,
               exit_on_legal: bool,
               exit_on_vulns: bool,
               werror: bool,
               vulns_confidence: str):
    ts_api = TrustSourceAPI(base_url, api_key)
    scans = load_scans_from_file(path, scan_format)

    if vulns_only:
        check_vulns(ts_api, scans, vulns_confidence, output_path, exit_on_vulns)
    else:
        if not project_name:
            print("Usage: ts-scan check [OPTIONS] PATH")
            print("Try 'ts-scan check --help' for help.")
            print("\nError: Missing option '--project-name'.")
            exit(2)
        check_comps(ts_api, scans, project_name, exit_on_legal, exit_on_vulns, werror, output_path)


def check_comps(api: TrustSourceAPI,
                scans: t.List[DependencyScan],
                proj_name: str,
                exit_on_legal: bool,
                exit_on_vulns: bool,
                werror: bool,
                output_path: t.Optional[Path]):

    res = {}

    for scan in scans:
        comps = []

        for dep in scan.iterdeps_once():
            for v in dep.versions:
                comps.append({
                    'key': dep.key,
                    'name': dep.name,
                    'version': v
                })

        if not comps:
            continue

        try:
            check = api.check_components(proj_name, scan.module, comps)
        except TrustSourceAPI.Error as err:
            print(f"An error occured while requesting vulnerabilities information for the scan")
            print(err)
            exit(2)

        if not check:
            continue

        res[scan.module] = check
        if output_path:
            with output_path.resolve().open('w') as fp:
                json.dump(res, fp, indent=2)  # noqa

        for w in check.get('warnings', []):
            comp = f'{w.get("component")}:{w.get("version")}'
            status = w.get('status').replace('_', ' ')
            msg.warn(f'{comp}: {status}')

        if exit_on_legal:
            violations = 0
            warnings = 0

            for res in check['data']:
                comp = res['component']
                comp_key = f'{comp["key"]}:{comp["version"]}'

                res = res['not_changed']

                for v in res['violations']:
                    log_msg = f'{comp_key}: {v["message"]}'
                    if v['type'] == 'violation':
                        msg.fail(log_msg)
                        violations += 1
                    elif v['type'] == 'warning':
                        msg.warn(log_msg)
                        warnings += 1

                    msg.divider(char=' ')

            if violations > 0 or (werror and warnings > 0):
                exit(1)

        if exit_on_vulns:
            violations = 0
            warnings = 0
            vulns = {}

            for res in check['data']:
                comp = res['component']
                comp_key = f'{comp["key"]}:{comp["version"]}'

                vulns.setdefault(comp_key, [])

                for v in res['vulnerabilities']:
                    vulns[comp_key].append(v["name"])
                    if v['status'] == 'violations':
                        violations += 1
                    elif v['status'] == 'warning':
                        warnings += 1

            for comp_key, vulns in vulns.items():
                if vulns:
                    msg.fail(f'{comp_key}:{vulns}')

            if violations > 0 or (werror and warnings > 0):
                exit(1)


def check_vulns(api: TrustSourceAPI,
                scans: t.List[DependencyScan],
                vulns_confidence: str,
                output_path: t.Optional[Path],
                exit_on_vulns: bool):

    if vulns_confidence == 'high':
        vulns_confidence = 1
    elif vulns_confidence == 'medium':
        vulns_confidence = 2
    else:
        vulns_confidence = 3

    vulns = {}

    for scan in scans:
        vulns.update(eval_vulns(scan, vulns_confidence, api))

    if not vulns:
        msg.good('No vulnerabilities identified')
    else:
        if output_path:
            with output_path.resolve().open('w') as fp:
                json.dump(vulns, fp, indent=2)  # noqa
        else:
            print(json.dumps(vulns, indent=2))

        if exit_on_vulns:
            exit(1)


def eval_vulns(scan: DependencyScan, confidence: int, api: TrustSourceAPI) -> t.Dict[str, dict]:
    vulns = {}
    purls = [dep.purl.to_string() for dep in scan.iterdeps_once()]
    pbar = tqdm(desc="Checking vulnerabilities", total=len(purls))

    def _check_completed(_task):
        result = _task.result()
        for res in result:
            comp_vulns = []
            for res_cves in res.get('cves', []):
                if (cves := res_cves['cves']) and cves and (res_cves.get('confidence', 0) < confidence):
                    comp_vulns.append(res_cves)

            if comp_vulns:
                vulns[res['key']] = comp_vulns

        pbar.update(len(result))

    def _check_vulns(_purls: t.List[str], _api: TrustSourceAPI) -> list:
        try:
            return _api.find_cves(_purls)

        except TrustSourceAPI.Error as err:
            print(f"An error occured while requesting vulnerabilities information for the scan")
            print(err)
            return []

    tasks = []
    pool = futures.ThreadPoolExecutor()
    chunk_size = min(len(purls) // pool._max_workers, 20)

    for i in range(0, len(purls), chunk_size):
        task = pool.submit(lambda _purls: _check_vulns(_purls, api), purls[i:i + chunk_size])
        task.add_done_callback(_check_completed)
        tasks.append(task)

    futures.wait(tasks, return_when=futures.ALL_COMPLETED)
    return vulns



import json
import typing as t

from pathlib import Path
from defusedxml import ElementTree
from copy import copy

from cyclonedx.model.bom import Bom
from cyclonedx.model.component import Component, ComponentType
from cyclonedx.model.license import LicenseExpression, DisjunctiveLicense
from cyclonedx.factory.license import LicenseFactory
from cyclonedx.schema import OutputFormat, SchemaVersion
from cyclonedx.output import make_outputter

from ..pm import Dependency, DependencyScan, License

_lic_factory = LicenseFactory()


def import_scan(path: Path, fmt: str) -> t.Optional[DependencyScan]:
    if fmt == 'cyclonedx-json':
        with path.open() as fp:
            bom_data = json.load(fp)
            bom = Bom.from_json(bom_data)

    elif fmt == 'cyclonedx-xml':
        with path.open() as fp:
            bom_data = ElementTree.fromstring(fp.read())
            bom = Bom.from_xml(bom_data)

    else:
        raise ValueError(f'Unsupported CycloneDX input format: {fmt}')

    return _create_scan(bom)


def export_scan(scan: DependencyScan, output: t.TextIO, fmt: str):
    if fmt == 'cyclonedx-json':
        output_fmt = OutputFormat.JSON
    elif fmt == 'cyclonedx-xml':
        output_fmt = OutputFormat.XML
    else:
        raise ValueError(f'Unsupported CycloneDX output format: {fmt}')

    bom = _create_bom(scan)
    outputter = make_outputter(bom, output_fmt, SchemaVersion.V1_6)

    output.write(outputter.output_as_string(indent=2))


def _create_bom(scan: DependencyScan) -> Bom:
    bom = Bom()
    bom.metadata.tools.components.add(Component(
        name='ts-scan',
        type=ComponentType.APPLICATION))

    bom.metadata.component = root = Component(
        name=scan.module
    )

    comps = [_create_component(dep, bom) for dep in scan.dependencies]
    bom.register_dependency(root, comps)

    return bom


def _create_component(dep: Dependency, bom: Bom) -> Component:
    comp = bom.get_component_by_purl(dep.purl)

    if not comp:
        comp = Component(
            name=dep.name,
            version=dep.version,
            licenses=[_lic_factory.make_from_string(lic.name) for lic in dep.licenses],
            purl=dep.purl
        )
        bom.components.add(comp)

    bom.register_dependency(comp, [_create_component(d, bom) for d in dep.dependencies])
    return comp


def _create_scan(bom: Bom) -> DependencyScan:
    deps = {}
    visited = {}

    for comp in bom.components:
        if dep := _create_dependency(comp):
            deps[comp.bom_ref] = dep

    for src_bom in bom.dependencies:
        if bom.metadata.component and src_bom.ref == bom.metadata.component.bom_ref:
            continue

        src = deps.get(src_bom.ref)
        if not src:
            src = visited.get(src_bom.ref)

        for dst_bom in src_bom.dependencies:
            if dst := deps.pop(dst_bom.ref, None):
                visited[dst_bom.ref] = dst
            else:
                dst = visited.get(dst_bom.ref)

            if src and dst:
                dst = copy(dst)
                dst.dependencies = []
                src.dependencies.append(dst)

    visited = {(dep.key, dep.version): dep for dep in visited.values()}

    stack = []
    stack.extend(deps.values())

    while len(stack) > 0:
        cur = stack.pop()
        cur_deps = []

        for dep in cur.dependencies:
            if d := visited.pop((dep.key, dep.version), None):
                cur_deps.append(d)
                stack.append(d)
            else:
                cur_deps.append(dep)

        cur.dependencies = cur_deps

    if bom.metadata.component:
        module = bom.metadata.component.name
    else:
        module = 'unknown'

    return DependencyScan(module=module, moduleId=f'cdx:{module}', dependencies=list(deps.values()))


def _create_dependency(comp: Component) -> t.Optional[Dependency]:
    if (purl := comp.purl) and (dep := Dependency.create_from_purl(purl)):
        dep.versions.append(comp.version if comp.version else purl.version)

        lics = []
        for lic in comp.licenses:
            if type(lic) is DisjunctiveLicense and lic.id or lic.name:
                lics.append(License(name=lic.id if lic.id else lic.name, url=lic.url))
            elif type(lic) is LicenseExpression and lic.value:
                lics.append(License(name=lic.value))

        dep.licenses = lics

        return dep

    return None



import json
import click
import requests

from pathlib import Path

from . import cli, api_default_options, sbom_formats


@cli.command('import', help='Imports SBOM documents directly to TrustSource API')
@click.option('-f', '--format', 'sbom_format', type=click.Choice(choices=list(sbom_formats.keys())), required=True,
              help='SBOM file format')
@click.option('-v', '--version', 'version', type=str, required=True, help='SBOM format version')
@click.option('--module', 'module', type=str, required=True, help='Module name')
@click.option('--module-id', 'moduleId', type=str, required=True, help='Module identifier')
@click.option('--pkg-type', 'pkgType', type=str, help='Type of pkg to apply for included components')
@click.option('--release', type=str, help='Release identifier')
@click.option('--branch', type=str, help='Code branch')
@click.option('--tag', type=str, help='Code tag')
@api_default_options
@click.argument('path', type=click.Path(exists=True, path_type=Path))
def import_sbom(sbom_format: str, **kwargs):
    path: Path = kwargs.pop('path')
    api_key: str = kwargs.pop('api_key')
    base_url: str = kwargs.pop('base_url')

    if sbom_format not in sbom_formats:
        print('Unsupported SBOM format')
        exit(1)

    url = f'{base_url}/core/imports/scan/{sbom_formats[sbom_format]}'

    headers = {
        'user-agent': 'ts-scan/1.0.0',
        'x-api-key': api_key
    }

    params = {k: v for k, v in kwargs.items() if v}

    with path.open('r') as fp:
        if sbom_format == 'cyclonedx':
            data = fp.read()
            headers['Content-Type'] = 'application/json'
            response = requests.post(url, data=data, headers=headers, params=params)
        else:
            content_type = 'application/json' if sbom_format == 'spdx-json' else 'application/xml'
            files = {'file': (path.name, fp, content_type)}
            headers['Content-Type'] = 'multipart/form-data'
            response = requests.post(url, files=files, headers=headers, params=params)

    if response.status_code == 201:
        print("Transfer success!")
        exit(0)
    else:
        print(json.dumps(response.text, indent=2))
        exit(1)



import re
import typing as t
import subprocess

from pathlib import Path
from tempfile import TemporaryDirectory

from xml.etree import ElementTree as ET

from .. import PackageManagerScanner, DependencyScan, Dependency

from .pom_utils import Pom
from .tree_utils import Tree


class MavenScanner(PackageManagerScanner):
    def __init__(self, excludeDepTypes: t.Optional[str] = None, **kwargs):
        super().__init__(**kwargs)

        self.__excludeDepTypes = excludeDepTypes.split(',') if excludeDepTypes else ['test']

        self.__path = None
        self.__nodes = None
        self.__processed_deps = set()

        self.__local_repo = None
        self.__remote_repos = {'central': 'https://repo.maven.apache.org/maven2'}

    @staticmethod
    def name() -> str:
        return 'Maven'

    @staticmethod
    def executable() -> t.Optional[str]:
        return 'mvn'

    @classmethod
    def options(cls) -> PackageManagerScanner.OptionsType:
        return super().options() | {
            'excludeDepTypes': {
                'type': str,
                'required': False,
                'help': 'A comma separated list of Maven dependency types to be excluded from the scan'
            }
        }

    def accepts(self, path: Path) -> bool:
        return ((path.is_dir() and (path / 'pom.xml').exists()) or
                (path.is_file() and path.name == 'pom.xml'))

    def scan(self, path: Path) -> t.Optional[DependencyScan]:
        return self._execute(path)

    def _execute(self, path: Path) -> t.Optional['DependencyScan']:
        self.__path = path

        with TemporaryDirectory() as temp_dir:
            temp_dir = Path(temp_dir)

            self.__local_repo = self._find_local_repository(temp_dir)
            self.__remote_repos.update(self._find_remote_repositories(temp_dir))

            tree_file = temp_dir / 'deps.tree'

            # Dump dependencies tree
            result = self._mvn('dependency:tree',
                               '-DoutputType=text', '-DappendOutput=true', f'-DoutputFile={tree_file}')

            if result.returncode != 0:
                print('Failed to dump dependency tree')
                exit(1)

            # Resolve dependencies sources
            self._mvn('dependency:sources', '-fn')

            scan = None
            if nodes := Tree.from_maven_file(tree_file):
                deps = []
                self.__nodes = nodes

                for n in nodes:
                    if dep := self._create_dep_from_node(n):
                        deps.append(dep)

                name = self._evaluate('project.name', temp_dir)
                groupId = self._evaluate('project.groupId', temp_dir)
                artifactId = self._evaluate('project.artifactId', temp_dir)
                version = self._evaluate('project.version', temp_dir)

                scan = DependencyScan(module=name,
                                      moduleId=f'mvn:{groupId}:{artifactId}' + f':{version}' if version else '',
                                      dependencies=deps)

            return scan

    def _mvn(self, *args, **kwargs) -> subprocess.CompletedProcess:
        return self._exec('-f', self.__path, *args, **kwargs)

    def _evaluate(self, expr: str, output: Path) -> t.Optional[str]:
        out = None
        out_file = output / 'eval.out'
        if self._mvn('help:3.3.0:evaluate', f'-Dexpression={expr}', f'-Doutput={out_file}').returncode == 0:
            with out_file.open('r') as fp:
                out = fp.read()
            out_file.unlink()
        return out

    def _create_dep_from_node(self, node: Tree) -> t.Optional[Dependency]:
        # example coordinates: org.tmatesoft.svnkit:svnkit:jar:1.8.7:provided

        try:
            group_id, artifact_id, _, version, *other = node.data.split(":")

            if len(other) > 0:
                if other[0] in self.__excludeDepTypes:
                    return None

            dep = MavenDependency(group_id=group_id,
                                  artifact_id=artifact_id,
                                  version=version,
                                  remote_repos=self.__remote_repos,
                                  local_repo=self.__local_repo)

            if artifact_id not in self.__processed_deps:
                dep.load()

                if pkg_data := dep.package_data:
                    _, _, checksum = pkg_data

                    if checksum:
                        dep.checksum = checksum[1]

                if src_data := dep.sources_data:
                    repo, pkg, checksum = src_data

                    dep.package_files.append(str(pkg))

                    download_url = f'{repo}/{pkg.relative_to(self.__local_repo)}'
                    sources_meta = {
                        'url': download_url
                    }

                    if checksum:
                        sources_meta['checksum'] = {
                            checksum[0]: checksum[1]
                        }

                    dep.meta['sources'] = sources_meta

                self.__processed_deps.add(artifact_id)

                for child in node.children:
                    if child_dep := self._create_dep_from_node(child):
                        dep.dependencies.append(child_dep)

            return dep

        except Exception as err:
            print(err)
            raise err

    def _get_project_modules(self, workdir: Path) -> t.List[str]:
        if res := self._evaluate('project.modules', workdir):
            try:
                return [node.text for node in ET.fromstring(res)]
            except:
                pass

        return []

    def _create_effective_pom(self, workdir: Path) -> t.Optional[Pom]:
        pom_file = workdir / 'effective-pom.xml'
        result = self._mvn('help:effective-pom', f'-Doutput={pom_file}')

        if result.returncode == 0:
            return Pom.from_file(pom_file)

        return None

    def _find_local_repository(self, workdir: Path) -> t.Optional[Path]:
        if res := self._evaluate('settings.localRepository', workdir):
            return Path(res)
        else:
            return None

    def _find_remote_repositories(self, workdir: Path) -> t.Dict[str, str]:
        if res := self._evaluate('project.repositories', workdir):
            try:
                repos = {}
                for repo in ET.fromstring(res):
                    _id, _url = None, None
                    for prop in repo:
                        if prop.tag == 'id':
                            _id = prop.text
                        elif prop.tag == 'url':
                            _url = prop.text

                        if _id and _url:
                            repos[_id] = _url
                            _id, _url = None, None
                            break

                return repos

            except:
                pass

        return {}


class MavenDependency(Dependency):
    def __init__(self, group_id: str, artifact_id: str, version: str,
                 remote_repos: t.Dict[str, str],
                 local_repo: t.Optional[Path] = None,
                 **kwargs):

        super().__init__(key=f'mvn:{group_id}:{artifact_id}',
                         name=artifact_id,
                         versions=[version],
                         type='maven',
                         namespace=group_id,
                         **kwargs)

        self.__remote_repos = remote_repos

        if local_repo:
            local_repo_path = local_repo / Path(*group_id.split('.')) / Path(*artifact_id.split('.')) / version
            if local_repo_path.exists():
                self.__local_repo_path = local_repo_path
            else:
                self.__local_repo_path = None

        if self.__local_repo_path and (pom_file := next(self.__local_repo_path.glob('*.pom'), None)):
            self.__pom_file = pom_file
        else:
            self.__pom_file = None

    def load(self):
        if self.__pom_file and (pom := Pom.from_file(self.__pom_file)):
            self.homepageUrl = pom.url
            self.description = pom.description
            self.licenses = pom.licenses

    @property
    def package_data(self) -> t.Optional[t.Tuple[str, Path, t.Optional[t.Tuple[str, str]]]]:
        return self._find_artifact_data('.jar')

    @property
    def sources_data(self) -> t.Optional[t.Tuple[str, Path, t.Optional[t.Tuple[str, str]]]]:
        return self._find_artifact_data('-sources.jar')

    def _find_artifact_data(self, artifact_name_suffix: str) -> t.Optional[
        t.Tuple[str, Path, t.Optional[
            t.Tuple[str, str]]]]:
        """
        Returns sources repository url, sources jar file and optionally its checksum
        :return: (sources_repo, sources_jar, (checksum_alg, checksum))
        """

        if not self.__local_repo_path:
            return None

        remote_repos = self.__local_repo_path / '_remote.repositories'
        if not remote_repos.exists():
            return None

        artifact_data_re = re.compile(rf"([\w.\-]*{artifact_name_suffix})>(.*)=")

        with remote_repos.open('r') as fp:
            for line in fp:
                if m := artifact_data_re.search(line):
                    fname = m.group(1)

                    repo = m.group(2) if m.group(2) else 'central'
                    if repo_url := self.__remote_repos.get(repo, None):
                        checksum = (f for f in self.__local_repo_path.glob(f'{fname}.*')
                                    if any(f.suffix.startswith(a) for a in ['.sha', '.md5']))

                        if checksum := next(checksum, None):
                            alg = checksum.suffix[1:]
                            with checksum.open('r') as checksum_fp:
                                checksum = alg, next(checksum_fp, None)

                        return repo_url, self.__local_repo_path / fname, checksum

        return None



import typing as t

from scanoss.scanossapi import ScanossApi
from scanoss.scanossgrpc import ScanossGrpc, DEFAULT_URL2

from tqdm import tqdm
from ts_deepscan import Scan as DSScan

from .. import __version__, msg
from ..pm import DependencyScan

__scanoss_api = None
__ts_scan_user_agent = f'ts-scan/{__version__}'


def get_scanoss_api() -> ScanossApi:
    global __scanoss_api

    if not __scanoss_api:
        __scanoss_api = ScanossApi()
        __scanoss_api.headers['User-Agent'] = __ts_scan_user_agent
        __scanoss_api.headers['user-agent'] = __ts_scan_user_agent

    return __scanoss_api


def extend_ds(ds: DSScan, api_key: t.Optional[str]):
    wfps = []
    for res in ds.result.values():
        if (scanoss := res.get('scanoss')) and (wfp := scanoss.get('wfp')):
            wfps.append(wfp)

    api = get_scanoss_api()

    linked_comps: t.Dict[str, t.Set[str]] = {}

    if wfps:
        wfps_results = {}
        wfps_chunk_size = 999

        for i in range(0, len(wfps), wfps_chunk_size):
            try:
                wfps_results.update(api.scan('\n'.join(wfps[i:i + wfps_chunk_size])))
            except:
                continue

        if wfps_results:
            for path, wfp_res in wfps_results.items():
                links = []
                for wfp_comp in wfp_res:
                    if purl := wfp_comp.get('purl'):
                        version = wfp_comp.get('version')

                        link = {'purl': purl}

                        if version:
                            link['version'] = version

                        if lics := wfp_comp.get('licenses'):
                            link['licenses'] = list(set(lic['name'] for lic in lics))

                        for p in purl:
                            p_vers = linked_comps.get(p, set())
                            if version:
                                p_vers.add(version)
                            linked_comps[p] = p_vers

                        links.append(link)

                if links:
                    ds.result[path]['links'] = links

    ds.summary['links'] = {purl: list(vers) for purl, vers in linked_comps.items()}


def analyse_scan(scan: DependencyScan, api_key: t.Optional[str]):
    for comp, ds in tqdm(scan.deepscans.items(), desc="Requesting fingerprints information from SCANOSS"):
        if 'links' not in ds.summary:
            extend_ds(ds, api_key)

    scan_purls = scan.as_purls_dict()

    if scan_purls and api_key:
        msg.info("Requesting cryptographic algorithms from SCANOSS")

        grpc_api = ScanossGrpc(url=DEFAULT_URL2, api_key=api_key)
        grpc_api.metadata.append(('user-agent', __ts_scan_user_agent))

        purls = {
            'purls': [{'purl': purl} for purl in scan_purls.keys()]
        }

        if crypto_info := grpc_api.get_crypto_json(purls=purls):
            for purl_crypto in crypto_info.get('purls', []):
                if (purl := purl_crypto.get('purl')) and (algorithms := purl_crypto.get('algorithms')):
                    if dep := scan_purls.get(purl):
                        for alg in algorithms:
                            dep.add_crypto_algorithm(algorithm=alg['algorithm'], strength=alg['strength'])




# generated by datamodel-codegen:
#   filename:  schema-16.0.18.json
#   timestamp: 2024-10-16T16:17:37+00:00

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from pydantic import AwareDatetime, BaseModel, Field, RootModel, constr


class CConanFileEntry(BaseModel):
    ref: str


class CConanInfoEntry(BaseModel):
    ref: str
    package_id: Optional[str] = None


class CConanLockV2Entry(BaseModel):
    ref: str
    packageID: Optional[str] = None
    username: Optional[str] = None
    channel: Optional[str] = None
    recipeRevision: Optional[str] = None
    packageRevision: Optional[str] = None
    timestamp: Optional[str] = None


class CPE(BaseModel):
    cpe: str
    source: Optional[str] = None


class CocoaPodfileLockEntry(BaseModel):
    checksum: str


class Coordinates(BaseModel):
    path: str
    layerID: Optional[str] = None


class DartPubspecLockEntry(BaseModel):
    name: str
    version: str
    hosted_url: Optional[str] = None
    vcs_url: Optional[str] = None


class Descriptor(BaseModel):
    name: str
    version: str
    configuration: Optional[Any] = None


class Digest(BaseModel):
    algorithm: str
    value: str


class DotnetDepsEntry(BaseModel):
    name: str
    version: str
    path: str
    sha512: str
    hashPath: str


class DotnetPortableExecutableEntry(BaseModel):
    assemblyVersion: str
    legalCopyright: str
    comments: Optional[str] = None
    internalName: Optional[str] = None
    companyName: str
    productName: str
    productVersion: str


class DpkgFileRecord(BaseModel):
    path: str
    digest: Optional[Digest] = None
    isConfigFile: bool


class ELFSecurityFeatures(BaseModel):
    symbolTableStripped: bool
    stackCanary: Optional[bool] = None
    nx: bool
    relRO: str
    pie: bool
    dso: bool
    safeStack: Optional[bool] = None
    cfi: Optional[bool] = None
    fortify: Optional[bool] = None


class ElfBinaryPackageNoteJsonPayload(BaseModel):
    type: Optional[str] = None
    architecture: Optional[str] = None
    osCPE: Optional[str] = None
    os: Optional[str] = None
    osVersion: Optional[str] = None
    system: Optional[str] = None
    vendor: Optional[str] = None
    sourceRepo: Optional[str] = None
    commit: Optional[str] = None


class ElixirMixLockEntry(BaseModel):
    name: str
    version: str
    pkgHash: str
    pkgHashExt: str


class ErlangRebarLockEntry(BaseModel):
    name: str
    version: str
    pkgHash: str
    pkgHashExt: str


class Executable(BaseModel):
    format: str
    hasExports: bool
    hasEntrypoint: bool
    importedLibraries: List[str]
    elfSecurityFeatures: Optional[ELFSecurityFeatures] = None


class FileLicenseEvidence(BaseModel):
    confidence: int
    offset: int
    extent: int


class FileMetadataEntry(BaseModel):
    mode: int
    type: str
    linkDestination: Optional[str] = None
    userID: int
    groupID: int
    mimeType: str
    size: int


class GoModuleEntry(BaseModel):
    h1Digest: Optional[str] = None


class HaskellHackageStackEntry(BaseModel):
    pkgHash: Optional[str] = None


class HaskellHackageStackLockEntry(BaseModel):
    pkgHash: Optional[str] = None
    snapshotURL: Optional[str] = None


class IDLikes(RootModel[List[str]]):
    root: List[str]


class JavaPomParent(BaseModel):
    groupId: str
    artifactId: str
    version: str


class JavaPomProject(BaseModel):
    path: str
    parent: Optional[JavaPomParent] = None
    groupId: str
    artifactId: str
    version: str
    name: str
    description: Optional[str] = None
    url: Optional[str] = None


class JavaPomProperties(BaseModel):
    path: str
    name: str
    groupId: str
    artifactId: str
    version: str
    scope: Optional[str] = None
    extraFields: Optional[Dict[constr(pattern=r'.*'), str]] = None


class JavaVMRelease(BaseModel):
    implementor: Optional[str] = None
    implementorVersion: Optional[str] = None
    javaRuntimeVersion: Optional[str] = None
    javaVersion: Optional[str] = None
    javaVersionDate: Optional[str] = None
    libc: Optional[str] = None
    modules: Optional[List[str]] = None
    osArch: Optional[str] = None
    osName: Optional[str] = None
    osVersion: Optional[str] = None
    source: Optional[str] = None
    buildSource: Optional[str] = None
    buildSourceRepo: Optional[str] = None
    sourceRepo: Optional[str] = None
    fullVersion: Optional[str] = None
    semanticVersion: Optional[str] = None
    buildInfo: Optional[str] = None
    jvmVariant: Optional[str] = None
    jvmVersion: Optional[str] = None
    imageType: Optional[str] = None
    buildType: Optional[str] = None


class JavascriptNpmPackage(BaseModel):
    name: str
    version: str
    author: str
    homepage: str
    description: str
    url: str
    private: bool


class JavascriptNpmPackageLockEntry(BaseModel):
    resolved: str
    integrity: str


class JavascriptYarnLockEntry(BaseModel):
    resolved: str
    integrity: str


class KeyValue(BaseModel):
    key: str
    value: str


class KeyValues(RootModel[List[KeyValue]]):
    root: List[KeyValue]


class LinuxKernelArchive(BaseModel):
    name: str
    architecture: str
    version: str
    extendedVersion: Optional[str] = None
    buildTime: Optional[str] = None
    author: Optional[str] = None
    format: Optional[str] = None
    rwRootFS: Optional[bool] = None
    swapDevice: Optional[int] = None
    rootDevice: Optional[int] = None
    videoMode: Optional[str] = None


class LinuxKernelModuleParameter(BaseModel):
    type: Optional[str] = None
    description: Optional[str] = None


class LinuxRelease(BaseModel):
    prettyName: Optional[str] = None
    name: Optional[str] = None
    id: Optional[str] = None
    idLike: Optional[IDLikes] = None
    version: Optional[str] = None
    versionID: Optional[str] = None
    versionCodename: Optional[str] = None
    buildID: Optional[str] = None
    imageID: Optional[str] = None
    imageVersion: Optional[str] = None
    variant: Optional[str] = None
    variantID: Optional[str] = None
    homeURL: Optional[str] = None
    supportURL: Optional[str] = None
    bugReportURL: Optional[str] = None
    privacyPolicyURL: Optional[str] = None
    cpeName: Optional[str] = None
    supportEnd: Optional[str] = None


class Location(BaseModel):
    path: str
    layerID: Optional[str] = None
    accessPath: str
    annotations: Optional[Dict[constr(pattern=r'.*'), str]] = None


class LuarocksPackage(BaseModel):
    name: str
    version: str
    license: str
    homepage: str
    description: str
    url: str
    dependencies: Dict[constr(pattern=r'.*'), str]


class MicrosoftKbPatch(BaseModel):
    product_id: str
    kb: str


class NixStoreEntry(BaseModel):
    outputHash: str
    output: Optional[str] = None
    files: List[str]


class OpamPackage(BaseModel):
    name: str
    version: str
    licenses: List[str]
    url: str
    checksum: List[str]
    homepage: str
    dependencies: List[str]


class PhpComposerAuthors(BaseModel):
    name: str
    email: Optional[str] = None
    homepage: Optional[str] = None


class PhpComposerExternalReference(BaseModel):
    type: str
    url: str
    reference: str
    shasum: Optional[str] = None


class PhpComposerInstalledEntry(BaseModel):
    name: str
    version: str
    source: PhpComposerExternalReference
    dist: PhpComposerExternalReference
    require: Optional[Dict[constr(pattern=r'.*'), str]] = None
    provide: Optional[Dict[constr(pattern=r'.*'), str]] = None
    require_dev: Optional[Dict[constr(pattern=r'.*'), str]] = Field(
        None, alias='require-dev'
    )
    suggest: Optional[Dict[constr(pattern=r'.*'), str]] = None
    license: Optional[List[str]] = None
    type: Optional[str] = None
    notification_url: Optional[str] = Field(None, alias='notification-url')
    bin: Optional[List[str]] = None
    authors: Optional[List[PhpComposerAuthors]] = None
    description: Optional[str] = None
    homepage: Optional[str] = None
    keywords: Optional[List[str]] = None
    time: Optional[str] = None


class PhpComposerLockEntry(BaseModel):
    name: str
    version: str
    source: PhpComposerExternalReference
    dist: PhpComposerExternalReference
    require: Optional[Dict[constr(pattern=r'.*'), str]] = None
    provide: Optional[Dict[constr(pattern=r'.*'), str]] = None
    require_dev: Optional[Dict[constr(pattern=r'.*'), str]] = Field(
        None, alias='require-dev'
    )
    suggest: Optional[Dict[constr(pattern=r'.*'), str]] = None
    license: Optional[List[str]] = None
    type: Optional[str] = None
    notification_url: Optional[str] = Field(None, alias='notification-url')
    bin: Optional[List[str]] = None
    authors: Optional[List[PhpComposerAuthors]] = None
    description: Optional[str] = None
    homepage: Optional[str] = None
    keywords: Optional[List[str]] = None
    time: Optional[str] = None


class PhpPeclEntry(BaseModel):
    name: str
    version: str
    license: Optional[List[str]] = None


class PortageFileRecord(BaseModel):
    path: str
    digest: Optional[Digest] = None


class PythonDirectURLOriginInfo(BaseModel):
    url: str
    commitId: Optional[str] = None
    vcs: Optional[str] = None


class PythonFileDigest(BaseModel):
    algorithm: str
    value: str


class PythonFileRecord(BaseModel):
    path: str
    digest: Optional[PythonFileDigest] = None
    size: Optional[str] = None


class PythonPackage(BaseModel):
    name: str
    version: str
    author: str
    authorEmail: str
    platform: str
    files: Optional[List[PythonFileRecord]] = None
    sitePackagesRootPath: str
    topLevelPackages: Optional[List[str]] = None
    directUrlOrigin: Optional[PythonDirectURLOriginInfo] = None
    requiresPython: Optional[str] = None
    requiresDist: Optional[List[str]] = None
    providesExtra: Optional[List[str]] = None


class PythonPipRequirementsEntry(BaseModel):
    name: str
    extras: Optional[List[str]] = None
    versionConstraint: str
    url: Optional[str] = None
    markers: Optional[str] = None


class PythonPipfileLockEntry(BaseModel):
    hashes: List[str]
    index: str


class PythonPoetryLockDependencyEntry(BaseModel):
    name: str
    version: str
    optional: bool
    markers: Optional[str] = None
    extras: Optional[List[str]] = None


class PythonPoetryLockExtraEntry(BaseModel):
    name: str
    dependencies: List[str]


class RDescription(BaseModel):
    title: Optional[str] = None
    description: Optional[str] = None
    author: Optional[str] = None
    maintainer: Optional[str] = None
    url: Optional[List[str]] = None
    repository: Optional[str] = None
    built: Optional[str] = None
    needsCompilation: Optional[bool] = None
    imports: Optional[List[str]] = None
    depends: Optional[List[str]] = None
    suggests: Optional[List[str]] = None


class Relationship(BaseModel):
    parent: str
    child: str
    type: str
    metadata: Optional[Any] = None


class RpmFileRecord(BaseModel):
    path: str
    mode: int
    size: int
    digest: Digest
    userName: str
    groupName: str
    flags: str


class RubyGemspec(BaseModel):
    name: str
    version: str
    files: Optional[List[str]] = None
    authors: Optional[List[str]] = None
    homepage: Optional[str] = None


class RustCargoAuditEntry(BaseModel):
    name: str
    version: str
    source: str


class RustCargoLockEntry(BaseModel):
    name: str
    version: str
    source: str
    checksum: str
    dependencies: List[str]


class Schema(BaseModel):
    version: str
    url: str


class Source(BaseModel):
    id: str
    name: str
    version: str
    type: str
    metadata: Any


class SwiftPackageManagerLockEntry(BaseModel):
    revision: str


class SwiplpackPackage(BaseModel):
    name: str
    version: str
    author: str
    authorEmail: str
    packager: str
    packagerEmail: str
    homepage: str
    dependencies: List[str]


class WordpressPluginEntry(BaseModel):
    pluginInstallDirectory: str
    author: Optional[str] = None
    authorUri: Optional[str] = None


class Cpes(RootModel[List[CPE]]):
    root: List[CPE]


class AlpmFileRecord(BaseModel):
    path: Optional[str] = None
    type: Optional[str] = None
    uid: Optional[str] = None
    gid: Optional[str] = None
    time: Optional[AwareDatetime] = None
    size: Optional[str] = None
    link: Optional[str] = None
    digest: Optional[List[Digest]] = None


class ApkFileRecord(BaseModel):
    path: str
    ownerUid: Optional[str] = None
    ownerGid: Optional[str] = None
    permissions: Optional[str] = None
    digest: Optional[Digest] = None


class CConanLockEntry(BaseModel):
    ref: str
    package_id: Optional[str] = None
    prev: Optional[str] = None
    requires: Optional[List[str]] = None
    build_requires: Optional[List[str]] = None
    py_requires: Optional[List[str]] = None
    options: Optional[KeyValues] = None
    path: Optional[str] = None
    context: Optional[str] = None


class ClassifierMatch(BaseModel):
    classifier: str
    location: Location


class DpkgDbEntry(BaseModel):
    package: str
    source: str
    version: str
    sourceVersion: str
    architecture: str
    maintainer: str
    installedSize: int
    provides: Optional[List[str]] = None
    depends: Optional[List[str]] = None
    preDepends: Optional[List[str]] = None
    files: List[DpkgFileRecord]


class FileLicense(BaseModel):
    value: str
    spdxExpression: str
    type: str
    evidence: Optional[FileLicenseEvidence] = None


class GoModuleBuildinfoEntry(BaseModel):
    goBuildSettings: Optional[KeyValues] = None
    goCompiledVersion: str
    architecture: str
    h1Digest: Optional[str] = None
    mainModule: Optional[str] = None
    goCryptoSettings: Optional[List[str]] = None
    goExperiments: Optional[List[str]] = None


class JavaJvmInstallation(BaseModel):
    release: JavaVMRelease
    files: List[str]


class JavaManifest(BaseModel):
    main: Optional[KeyValues] = None
    sections: Optional[List[KeyValues]] = None


class License(BaseModel):
    value: str
    spdxExpression: str
    type: str
    urls: List[str]
    locations: List[Location]


class LinuxKernelModule(BaseModel):
    name: Optional[str] = None
    version: Optional[str] = None
    sourceVersion: Optional[str] = None
    path: Optional[str] = None
    description: Optional[str] = None
    author: Optional[str] = None
    license: Optional[str] = None
    kernelVersion: Optional[str] = None
    versionMagic: Optional[str] = None
    parameters: Optional[Dict[constr(pattern=r'.*'), LinuxKernelModuleParameter]] = None


class PortageDbEntry(BaseModel):
    installedSize: int
    files: List[PortageFileRecord]


class PythonPoetryLockEntry(BaseModel):
    index: str
    dependencies: List[PythonPoetryLockDependencyEntry]
    extras: Optional[List[PythonPoetryLockExtraEntry]] = None


class RpmArchive(BaseModel):
    name: str
    version: str
    epoch: Optional[int] = None
    architecture: str
    release: str
    sourceRpm: str
    size: int
    vendor: str
    modularityLabel: Optional[str] = None
    provides: Optional[List[str]] = None
    requires: Optional[List[str]] = None
    files: List[RpmFileRecord]


class RpmDbEntry(BaseModel):
    name: str
    version: str
    epoch: Optional[int] = None
    architecture: str
    release: str
    sourceRpm: str
    size: int
    vendor: str
    modularityLabel: Optional[str] = None
    provides: Optional[List[str]] = None
    requires: Optional[List[str]] = None
    files: List[RpmFileRecord]


class Licenses(RootModel[List[License]]):
    root: List[License]


class AlpmDbEntry(BaseModel):
    basepackage: str
    package: str
    version: str
    description: str
    architecture: str
    size: int
    packager: str
    url: str
    validation: str
    reason: int
    files: List[AlpmFileRecord]
    backup: List[AlpmFileRecord]
    provides: Optional[List[str]] = None
    depends: Optional[List[str]] = None


class ApkDbEntry(BaseModel):
    package: str
    originPackage: str
    maintainer: str
    version: str
    architecture: str
    url: str
    description: str
    size: int
    installedSize: int
    pullDependencies: List[str]
    provides: List[str]
    pullChecksum: str
    gitCommitOfApkPort: str
    files: List[ApkFileRecord]


class BinarySignature(BaseModel):
    matches: List[ClassifierMatch]


class File(BaseModel):
    id: str
    location: Coordinates
    metadata: Optional[FileMetadataEntry] = None
    contents: Optional[str] = None
    digests: Optional[List[Digest]] = None
    licenses: Optional[List[FileLicense]] = None
    executable: Optional[Executable] = None
    unknowns: Optional[List[str]] = None


class JavaArchive(BaseModel):
    virtualPath: str
    manifest: Optional[JavaManifest] = None
    pomProperties: Optional[JavaPomProperties] = None
    pomProject: Optional[JavaPomProject] = None
    digest: Optional[List[Digest]] = None


class Package(BaseModel):
    id: str
    name: str
    version: str
    type: str
    foundBy: str
    locations: List[Location]
    licenses: Licenses
    language: str
    cpes: Cpes
    purl: str
    metadataType: Optional[str] = None
    metadata: Optional[
        Union[
            AlpmDbEntry,
            ApkDbEntry,
            BinarySignature,
            CConanFileEntry,
            CConanInfoEntry,
            CConanLockEntry,
            CConanLockV2Entry,
            CocoaPodfileLockEntry,
            DartPubspecLockEntry,
            DotnetDepsEntry,
            DotnetPortableExecutableEntry,
            DpkgDbEntry,
            ElfBinaryPackageNoteJsonPayload,
            ElixirMixLockEntry,
            ErlangRebarLockEntry,
            GoModuleBuildinfoEntry,
            GoModuleEntry,
            HaskellHackageStackEntry,
            HaskellHackageStackLockEntry,
            JavaArchive,
            JavaJvmInstallation,
            JavascriptNpmPackage,
            JavascriptNpmPackageLockEntry,
            JavascriptYarnLockEntry,
            LinuxKernelArchive,
            LinuxKernelModule,
            LuarocksPackage,
            MicrosoftKbPatch,
            NixStoreEntry,
            OpamPackage,
            PhpComposerInstalledEntry,
            PhpComposerLockEntry,
            PhpPeclEntry,
            PortageDbEntry,
            PythonPackage,
            PythonPipRequirementsEntry,
            PythonPipfileLockEntry,
            PythonPoetryLockEntry,
            RDescription,
            RpmArchive,
            RpmDbEntry,
            RubyGemspec,
            RustCargoAuditEntry,
            RustCargoLockEntry,
            SwiftPackageManagerLockEntry,
            SwiplpackPackage,
            WordpressPluginEntry,
        ]
    ] = None


class Document(BaseModel):
    artifacts: List[Package]
    artifactRelationships: List[Relationship]
    files: Optional[List[File]] = None
    source: Source
    distro: LinuxRelease
    descriptor: Descriptor
    schema_: Schema = Field(..., alias='schema')


class Model(RootModel[Document]):
    root: Document



import typing as t
import tempfile

from pathlib import Path
from urllib.parse import urlparse
from packageurl import PackageURL

from .model import Document, Package
from ts_scan.pm import Scanner, Dependency, DependencyScan, License
from ts_scan.cli import msg


class SyftScanner(Scanner):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def name() -> str:
        return "Syft"

    @staticmethod
    def executable() -> t.Optional[str]:
        return 'syft'

    def accepts(self, path: Path) -> bool:
        return True

    def scan(self, src: t.Union[str, Path]) -> t.Optional[DependencyScan]:
        with tempfile.TemporaryDirectory() as tmpdir:
            output = Path(f'{tmpdir}/scan.json')

            res = self._exec(str(src), '-o', f'syft-json={output}')

            if res.returncode != 0:
                msg.fail(f'Syft failed to scan {str(src)}. Error code: {res.returncode}')
                return None

            if scan := self.__create_scan(output):
                if isinstance(src, Path):
                    scan.moduleId = f'{src.name}:{scan.module}'
                else:
                    try:
                        url = urlparse(src)
                        scan.moduleId = f'{url.scheme}:{scan.module}'
                    except ValueError:
                        pass

                return scan

    @staticmethod
    def __create_scan(path: Path) -> t.Optional[DependencyScan]:
        with path.open() as fp:
            data = fp.read()

        doc = Document.model_validate_json(data)

        if distro := doc.distro:
            module = distro.id
        else:
            module = doc.descriptor.name

        deps = [SyftScanner.__create_dep(pkg) for pkg in doc.artifacts]

        return DependencyScan(module=module, moduleId=module, dependencies=deps)

    @staticmethod
    def __create_dep(pkg: Package, use_purl_as_version=False) -> t.Optional[Dependency]:
        dep = None

        try:
            purl = PackageURL.from_string(pkg.purl)
        except ValueError:
            purl = None

        if purl:
            key = SyftScanner._map_purl_type(purl.type)
            if purl.namespace:
                key += ':' + purl.namespace
            key += ':' + purl.name

            if not use_purl_as_version:
                ver = pkg.version
            else:
                ver = pkg.purl

            ns = purl.namespace if purl.namespace else ''

            dep = Dependency(key, pkg.name, versions=[ver], type=purl.type, namespace=ns)
            dep.meta['purl'] = pkg.purl

            for pkg_lic in pkg.licenses.root:
                if pkg_lic.type == 'declared':
                    lic = License(pkg_lic.value)
                    if pkg_lic.urls:
                        lic.url = pkg_lic.urls[0]
                    dep.licenses.append(lic)

        return dep

    @staticmethod
    def _map_purl_type(ty: str):
        # TrustSource key mapping
        if ty == 'maven':
            return 'mvn'
        else:
            return ty



import click
import typing as t
import ts_deepscan

from pathlib import Path

from . import cli, load_scans_from_file
from .scan import output_scans
from .. import DependencyScan

from ..analyse import analyse_scan_with_ds, analyse_path_with_ds
from ..analyse.scanoss import analyse_scan as analyse_scan_with_scanoss


@cli.command('analyse', help='Analyze scanned dependencies or folder contents')
@cli.inout_default_options(_in=True, _out=True, _fmt=True)
@click.option('--disable-deepscan', default=False, is_flag=True,
              help='Disable scanning of the package\'s sources if available using TrustSource Deepscan')
@click.option('--disable-scanoss', default=False, is_flag=True,
              help='Disable analysing results using ScanOSS scanner')
@click.option('--scanoss-api-key', 'scanoss_api_key', type=str, required=False, help='SCANOSS API Key')
@click.option('--Xdeepscan', default=[], multiple=True,
              help='Specifies an option which should be passed to the TrustSource DeepScan')
def analyse_scan(path: Path,
                 output_path: t.Optional[Path],
                 scan_format: str,
                 disable_deepscan: bool,
                 disable_scanoss: bool,
                 scanoss_api_key: t.Optional[str],
                 xdeepscan: tuple[str]):

    xdeepscan = list(xdeepscan)

    if not disable_deepscan:
        if not disable_scanoss and "--include-scanoss-wfp" not in xdeepscan:
            xdeepscan.append("--include-scanoss-wfp")

    analysed_scans = []

    if path.is_dir():
        if not disable_deepscan:
            ds_scan = analyse_path_with_ds(path, ds_args=xdeepscan)
            scan = DependencyScan(module='unknown', moduleId='unknown')
            scan.deepscans['unknown'] = ds_scan

            if not disable_scanoss:
                analyse_scan_with_scanoss(scan, scanoss_api_key)

            analysed_scans.append(scan)
    else:
        scans = load_scans_from_file(path, scan_format)

        for s in scans:
            # Apply DS analysis
            if not disable_deepscan:
                analyse_scan_with_ds(s, ds_args=xdeepscan)

            # Apply ScanOSS analysis
            if not disable_scanoss:
                analyse_scan_with_scanoss(s, scanoss_api_key)

            analysed_scans.append(s)

    output_scans(analysed_scans, output_path)






import json
import ts_deepscan

from tqdm import tqdm
from pathlib import Path

from . import cli, load_scans_from_file
from ..api import TrustSourceAPI


@cli.command('upload', help='Transfers scan and analyse results to TrustSource API')
@cli.inout_default_options(_in=True, _out=False, _fmt=True)
@cli.api_default_options()
def upload_scan(path: Path,
                scan_format: str,
                project_name: str,
                base_url: str,
                api_key: str):

    api = TrustSourceAPI(base_url, api_key)
    ds_base_url = f'{base_url}/v2/repository'

    def _do_upload(data: dict):
        deepscans = data.pop('deepscans', {})

        if deepscans := [(k, ds) for k, ds in deepscans.items() if ds['stats']['total'] > 0]:
            deepscans_uploaded = {}

            for k, ds in tqdm(deepscans, desc='Uploading deepscans'):
                if res := ts_deepscan.upload_data(data=ds, module_name=k, api_key=api_key, base_url=ds_base_url):
                    deepscans_uploaded[k] = res

            deps = data.get('dependencies', []).copy()
            while deps:
                dep = deps.pop()
                if ds := deepscans_uploaded.get(dep['key']):
                    if versions := dep.get('versions', []):
                        ver = versions[0]
                    else:
                        ver = 'unknown'

                    meta = dep.get('meta', {})
                    meta['versions'] = [{
                        'version': ver,
                        'deepScanId': ds[0],
                        'deepScanUrl': ds[1]
                    }]
                    dep['meta'] = meta

                deps.extend(dep.get('dependencies', []))

        print('Uploading dependencies scan...')

        try:
            resp = api.upload_scan(data)
            print("Transfer success!")
            print(json.dumps(resp, indent=2))
        except TrustSourceAPI.Error as err:
            print("Transfer failed")
            print(err)

    ######

    scans = load_scans_from_file(path, scan_format)

    for s in scans:
        # noinspection PyUnresolvedReferences
        d = s.to_dict()
        d['project'] = project_name

        _do_upload(d)



# SPDX-FileCopyrightText: 2023 EACG GmbH
#
# SPDX-License-Identifier: Apache-2.0

import click
import typing as t

from wasabi import Printer
from pathlib import Path

from ..pm import DependencyScan, load_scans

msg = Printer(line_max=240)


def start():
    import ts_scan.cli.scan
    import ts_scan.cli.analyse
    import ts_scan.cli.check
    import ts_scan.cli.upload
    import ts_scan.cli.import_sbom
    import ts_scan.cli.convert

    cli()


@click.group()
@click.version_option(package_name='ts-scan')
def cli():
    pass


def api_default_options(project_name=True, is_project_name_required=True, project_name_help_hint=''):
    def _apply(f):
        if project_name:
            help_msg = 'Project name'
            if project_name_help_hint:
                help_msg += f' [{project_name_help_hint}]'
            f = click.option('--project-name', 'project_name',
                             type=str, required=is_project_name_required, help=help_msg)(f)

        f = click.option('--api-key', 'api_key',
                         type=str, required=True, help='TrustSource API Key')(f)
        f = click.option('--base-url', 'base_url',
                         default='https://api.trustsource.io', help='TrustSource API base URL')(f)
        return f

    return _apply


def inout_default_options(_in: bool, _out: bool, _fmt: bool):
    def _apply(f):
        if _in:
            f = click.argument('path', type=click.Path(exists=True, path_type=Path))(f)
        if _out:
            f = click.option('-o', '--output', 'output_path', required=False, type=click.Path(path_type=Path),
                             help='Output path for the scan')(f)
        if _fmt:
            f = click.option('-f', '--format', 'scan_format', type=click.Choice(choices=scan_formats),
                             default='ts', help='Scans file format')(f)
        return f

    return _apply


def parse_cmd_opts_from_args(cmd: click.Command, args: [str]):
    ctx = cmd.context_class(cmd)
    with ctx:
        parser = cmd.make_parser(ctx)
        params, _, param_order = parser.parse_args(args=args)

        opts = {}
        for param in param_order:
            if isinstance(param, click.Option):
                value, _ = param.handle_parse_result(ctx, params, [])
                opts[param.name] = value

    return opts


def load_scans_from_file(path: Path, scan_format: str) -> t.List[DependencyScan]:
    try:
        return load_scans(path, scan_format)
    except:
        msg.fail(f"Failed to load scan in the '{scan_format}' format. Please ensure the format is correct.")
        exit(2)


cli.api_default_options = api_default_options
cli.inout_default_options = inout_default_options

sbom_formats = {
    'spdx-rdf': 'spdx/rdf',
    'spdx-json': 'spdx/json',
    'cyclonedx': 'cyclonedx'
}

scan_formats = [
    'ts',
    'spdx-tag', 'spdx-json', 'spdx-yaml', 'spdx-xml',
    'cyclonedx-json', 'cyclonedx-xml'
]



import abc
import json
import click
import typing as t
import shutil
import subprocess

from abc import ABC
from enum import Enum
from sys import platform
from pathlib import Path
from packageurl import PackageURL

from dataclasses import dataclass, field
from dataclasses_json import config, dataclass_json
from typing_extensions import TextIO

from ts_deepscan.scanner import Scan as DSScan


class ExecutableNotFoundError(Exception):
    pass


class Scanner(abc.ABC):
    OptionsType = t.Dict[str, t.Dict[str, t.Any]]

    def __init__(self, verbose=False, ignore=False, executable: t.Optional[Path] = None,
                 forward: t.Optional[tuple] = None):
        self.verbose = verbose
        self.ignore = ignore
        self.executable_path = executable

        self.__forward = [arg for fwd in forward for arg in fwd.split(',')] if forward else []

    @staticmethod
    @abc.abstractmethod
    def name() -> str:
        raise NotImplemented()

    @staticmethod
    def executable() -> t.Optional[str]:
        return None

    @classmethod
    def options(cls) -> OptionsType:
        opts = {}

        if cls.executable():
            opts['executable'] = {
                'type': click.Path(path_type=Path),
                'required': False,
                'help': f'A path to the {cls.name()} executable'
            }
            opts['forward'] = {
                'type': str,
                'required': False,
                'multiple': True,
                'help': f'Forward parameters to the {cls.name()} executable'
            }

        return opts

    @abc.abstractmethod
    def accepts(self, path: Path) -> bool:
        raise NotImplemented()

    @abc.abstractmethod
    def scan(self, src: t.Union[str, Path]) -> t.Optional['DependencyScan']:
        raise NotImplemented()

    def _exec(self, *args, capture_output=False, **kwargs) -> subprocess.CompletedProcess:
        exec_path = self.executable_path if self.executable_path else self.__class__.executable()

        if cmd := shutil.which(exec_path):
            return subprocess.run(
                [cmd] + list(args) + self.__forward,
                shell=(platform == 'win32'),
                check=True,
                capture_output=capture_output or not self.verbose,
                **kwargs
            )

        else:
            raise ExecutableNotFoundError(f'Cannot find {exec_path} executable.')


class PackageManagerScanner(Scanner, ABC):
    OptionsType = t.Dict[str, t.Dict[str, t.Any]]

    @classmethod
    def options(cls) -> OptionsType:
        return super().options() | {
            'ignore': {
                'default': False,
                'is_flag': True,
                'help': f'Ignores scanning {cls.name()} dependencies'
            }
        }


@dataclass_json
@dataclass
class DependencyScan:
    module: str
    moduleId: str
    dependencies: t.List['Dependency'] = field(default_factory=lambda: [])

    tag: t.Optional[str] = field(default=None, metadata=config(exclude=lambda f: f is None))
    branch: t.Optional[str] = field(default=None, metadata=config(exclude=lambda f: f is None))

    deepscans: t.Dict[str, DSScan] = field(default_factory=lambda: {})

    def iterdeps(self) -> t.Iterable['Dependency']:
        deps = list(self.dependencies)
        while deps:
            d = deps.pop()
            deps.extend(d.dependencies)
            yield d

    def iterdeps_once(self) -> t.Iterable['Dependency']:
        visited = set()
        for dep in self.iterdeps():
            if dep.purl not in visited:
                visited.add(dep.purl)
                yield dep

    def as_purls_dict(self) -> t.Dict[str, 'Dependency']:
        return {dep.purl.to_string(): dep for dep in self.iterdeps_once()}


@dataclass_json
@dataclass
class Dependency:
    key: str
    name: str
    type: str

    namespace: str = ''

    repoUrl: str = ''
    homepageUrl: str = ''
    description: str = ''
    checksum: str = ''
    private: bool = False

    versions: t.List[str] = field(default_factory=lambda: [])
    dependencies: t.List['Dependency'] = field(default_factory=lambda: [])
    licenses: t.List['License'] = field(default_factory=lambda: [])

    meta: t.Dict = field(default_factory=lambda: {})

    package_files: t.List[str] = field(default_factory=lambda: [])
    license_file: t.Optional[str] = None

    crypto_algorithms: t.List['CryptoAlgorithm'] = field(default_factory=lambda: [])

    @property
    def version(self) -> t.Optional[str]:
        return self.versions[0] if len(self.versions) == 1 else None

    @property
    def purl(self):
        return PackageURL(type=self.type,
                          namespace=self.namespace,
                          name=self.name,
                          version=self.version)

    @staticmethod
    def create_from_purl(purl: t.Union[str, PackageURL]) -> t.Optional['Dependency']:
        if type(purl) is str:
            try:
                purl = PackageURL.from_string(purl)
            except ValueError:
                return None

        key = purl.type
        if purl.namespace:
            key += ':' + purl.namespace
        key += ':' + purl.name

        return Dependency(key=key,
                          name=purl.name,
                          type=purl.type,
                          namespace=purl.namespace,
                          meta={'purl': purl.to_string()})

    def add_crypto_algorithm(self, algorithm: str, strength: str):
        if next((a for a in self.crypto_algorithms if
                 a.algorithm == algorithm and a.strength == strength), None) is None:
            self.crypto_algorithms.append(CryptoAlgorithm(algorithm=algorithm, strength=strength))


class LicenseKind(str, Enum):
    DECLARED = 'declared'
    EFFECTIVE = 'effective'


@dataclass_json
@dataclass
class License:
    name: str
    url: str = ''
    kind: LicenseKind = LicenseKind.DECLARED


@dataclass_json
@dataclass
class CryptoAlgorithm:
    algorithm: str
    strength: str


def dump_scans(scans: t.List[DependencyScan], fp: TextIO, fmt: str):
    # noinspection PyProtectedMember
    from dataclasses_json.core import _ExtendedEncoder

    if not scans:
        return

    if fmt == 'ts':
        # noinspection PyUnresolvedReferences
        scans = [s.to_dict() for s in scans]
        # noinspection PyTypeChecker
        json.dump(scans, fp, cls=_ExtendedEncoder, indent=2)

    elif fmt in ['spdx-tag', 'spdx-json', 'spdx-yaml', 'spdx-xml']:
        from ..spdx import export_scan
        export_scan(scans[0], fp, fmt)

    elif fmt in ['cyclonedx-json', 'cyclonedx-xml']:
        from ..cyclonedx import export_scan
        export_scan(scans[0], fp, fmt)

    else:
        raise ValueError(f'Unsupported output format: {fmt}')


def load_scans(path: Path, fmt: str) -> t.List[DependencyScan]:
    if fmt == 'ts':
        with path.open('r') as fp:
            data = json.load(fp)
            if type(data) is list:
                # noinspection PyUnresolvedReferences
                return [DependencyScan.from_dict(d) for d in data]
            else:
                # noinspection PyUnresolvedReferences
                return [DependencyScan.from_dict(data)]

    elif fmt in ['spdx-tag', 'spdx-json', 'spdx-yaml', 'spdx-xml']:
        from ..spdx import import_scan
        return [import_scan(path, fmt)]

    elif fmt in ['cyclonedx-json', 'cyclonedx-xml']:
        from ..cyclonedx import import_scan
        return [import_scan(path, fmt)]

    else:
        raise ValueError(f'Unsupported input format: {fmt}')



import itertools
import typing as t
import importlib.resources
import concurrent.futures as futures

import ts_deepscan.cli
import ts_deepscan.analyser.textutils

from ts_deepscan import Scan as DSScan

from tqdm import tqdm
from pathlib import Path
from threading import Lock

from ..pm import DependencyScan, Dependency
from ..cli import parse_cmd_opts_from_args

__ds_dataset = None

__gitignore_patterns = {
    "npm": ["Node.gitignore"],
    "pypi": ["Python.gitignore"]
}


def _parse_ds_args(ds_args: t.List[str]) -> t.Tuple[t.List[str], t.Dict[str, t.Any]]:
    ds_args = list(itertools.chain.from_iterable(xd.split(',') for xd in ds_args))
    ds_opts = parse_cmd_opts_from_args(ts_deepscan.cli.scan, ds_args)  # noqa

    return ds_args, ds_opts


def _get_ds_dataset():
    global __ds_dataset

    if not __ds_dataset:
        __ds_dataset = ts_deepscan.create_dataset()

    return __ds_dataset


def _analyse_dep_with_ds(dep, dataset, **ds_opts) -> tuple:
    gitignores = [importlib.resources.path(f'{__package__}.gitignore', pat)
                  for pat in __gitignore_patterns.get(dep.type, [])]

    scanner = ts_deepscan.create_scanner(**ds_opts, default_gitignores=gitignores, dataset=dataset)  # noqa

    ds_res = None
    lic_file_res = None

    if dep.package_files:
        sources = [Path(src) for src in dep.package_files]
        ds_res = ts_deepscan.execute_scan(sources, scanner, title=dep.key)

    if (lic_file := dep.license_file) and __ds_dataset:
        lic_file_path = Path(lic_file)
        if lic_file_path.exists():
            with lic_file_path.open(errors="surrogateescape") as fp:
                lic_file_res = ts_deepscan.analyser.textutils.analyse_license_text(fp.read(), __ds_dataset)  # noqa

    return ds_res, lic_file_res


def _analyse_dep_with_ds_completed(dep, scan, pbar, completion_lock):
    def complete(task):
        ds_res, lic_file_res = task.result()

        with completion_lock:
            if ds_res:
                scan.deepscans[dep.key] = ds_res
                extend_dep_from_ds(dep, ds_res)

            if lic_file_res:
                dep.meta['license_file'] = lic_file_res

        with tqdm.get_lock():
            pbar.update()

    return complete


def analyse_scan_with_ds(scan: DependencyScan, ds_args: t.List[str]):
    if scan.deepscans:
        return scan

    ds_args, ds_opts = _parse_ds_args(ds_args)
    dataset = _get_ds_dataset()

    tasks = []
    pool = futures.ThreadPoolExecutor(max_workers=2)
    completion_lock = Lock()

    deps = [dep for dep in scan.iterdeps() if dep.package_files or dep.license_file]
    pbar = tqdm(desc="Analysing dependencies", total=len(deps))

    for dep in deps:
        task = pool.submit(_analyse_dep_with_ds, dep, dataset, **ds_opts)
        task.add_done_callback(_analyse_dep_with_ds_completed(dep, scan, pbar, completion_lock))
        tasks.append(task)

    futures.wait(tasks, return_when=futures.ALL_COMPLETED)

    pbar.close()


def analyse_path_with_ds(path: Path, ds_args: t.List[str]) -> DSScan:
    ds_args, ds_opts = _parse_ds_args(ds_args)
    dataset = _get_ds_dataset()

    scanner = ts_deepscan.create_scanner(**ds_opts, dataset=dataset)  # noqa

    return ts_deepscan.execute_scan([path], scanner, title=f"'{path.name}'")


def extend_dep_from_ds(dep: Dependency, ds: DSScan):
    from ..pm import License, LicenseKind

    for ds_lic in ds.summary.get('licenses', []):
        if next((lic for lic in dep.licenses
                 if lic.kind == LicenseKind.EFFECTIVE and lic.name == ds_lic), None) is None:
            dep.licenses.append(License(name=ds_lic, kind=LicenseKind.EFFECTIVE))

    for name, coding in ds.summary.get('crypto_algorithms', {}).items():
        dep.add_crypto_algorithm(algorithm=name, strength=coding)



# SPDX-FileCopyrightText: 2020 EACG GmbH
#
# SPDX-License-Identifier: Apache-2.0

import re
import build.util
import typing as t

from pathlib import Path
from importlib.metadata import distribution, PackageNotFoundError
from importlib.metadata._meta import PackageMetadata
from shippinglabel.requirements import parse_requirements

from . import PackageManagerScanner, DependencyScan, Dependency, License

_supported_pkg_files = [
    'setup.py',
    'pyproject.toml'
]


class PypiScanner(PackageManagerScanner):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.__processed_deps = set()

    @staticmethod
    def name() -> str:
        return "PyPI"

    def accepts(self, path: Path) -> bool:
        return path.is_dir() and any((path / pkg_file).exists() for pkg_file in _supported_pkg_files)

    def scan(self, path: Path) -> t.Optional[DependencyScan]:
        try:
            metadata = build.util.project_wheel_metadata(path, isolated=False)
        except:
            print(f'An error occured while building packages metadata')
            return None

        if dep := self._create_dep_from_metadata(metadata):
            return DependencyScan(module=dep.name, moduleId=f'pip:{dep.name}', dependencies=[dep])
        else:
            return None

    def _create_dep_from_metadata(self, metadata: PackageMetadata) -> Dependency:
        """
        Creates a dependency from the dist package metadate
        :param metadata: PackageMetadata
        :return:
        """
        name = metadata.get('Name', '')
        key = 'pip:' + name.lower()

        dep = Dependency(key=key, name=name, type='pypi')

        if version := metadata.get('Version', None):
            dep.versions.append(version)

        if name not in self.__processed_deps:
            self.__processed_deps.add(name)

            if licence := metadata.get('License', None):
                dep.licenses.append(License(name=licence))

            dep.repoUrl = metadata.get('Download-URL', '')
            dep.description = metadata.get('Summary', '')
            dep.homepageUrl = metadata.get('Home-page', '')

            dist = distribution(name)

            if lic_file := metadata.get('License-File', None):
                # noinspection PyTypeChecker
                dep.license_file = str(Path(dist.locate_file(lic_file)).resolve())

            if top_level := dist.read_text('top_level.txt'):
                # noinspection PyTypeChecker
                files = (str(Path(dist.locate_file(f)).resolve()) for f in top_level.split('\n') if f)
                dep.package_files.extend(files)

            # reqs = metadata.get_all('Requires-Dist', [])
            if reqs := dist.requires:
                req_pkgs = list(_extract_required_pkgs(reqs))
                dep.dependencies = [d for pkg in req_pkgs if (d := self._create_dep(pkg))]

        return dep

    def _create_dep(self, pkg: str) -> t.Optional[Dependency]:
        try:
            pkg_info = distribution(pkg)
        except PackageNotFoundError:
            return None

        return self._create_dep_from_metadata(pkg_info.metadata)


_import_statement_regex = re.compile(r'(?:from|import) ([A-Z0-9_]+).*', flags=re.IGNORECASE)


def _extract_imported_pkgs(path: Path) -> t.List[str]:
    with path.open('r') as fp:
        try:
            data = fp.read()
            return _import_statement_regex.findall(data)
        except:
            return []


def _extract_required_pkgs(reqs: t.List[str]) -> t.Iterable[str]:
    reqs, _ = parse_requirements(reqs)
    return {req.name for req in reqs}



import typing as t
from io import StringIO

import click
import click_params

from pathlib import Path
from urllib.parse import urlparse

from . import cli
from .. import (msg,
                do_scan,
                do_scan_with_syft,
                process_scan)

from ..pm import dump_scans, ExecutableNotFoundError


@cli.command('scan', help='Scans a target and determines its composition')
@cli.inout_default_options(_in=False, _out=True, _fmt=True)
@cli.scanner_options
@click.option('--verbose', default=False, is_flag=True,
              help="Verbose mode")
@click.option('--tag', required=False, type=str,
              help="Project's tag in the VCS")
@click.option('--branch', required=False, type=str,
              help="Project's branch in the VCS")
@click.option('--use-syft', default=False, is_flag=True,
              help='Use Syft scanner for the file system scan')
@click.argument('sources', type=click_params.FirstOf(click.Path(exists=True, path_type=Path), click.STRING), nargs=-1)
def scan_dependencies(sources: t.List[t.Union[Path, str]],
                      output_path: t.Optional[Path],
                      scan_format: str,
                      verbose: bool,
                      tag: str,
                      branch: str,
                      use_syft: bool,
                      **kwargs):

    def _do_scan():
        try:
            if use_syft:
                yield from do_scan_with_syft(sources, verbose=verbose, **kwargs)
            else:
                paths = []
                urls = []
                for src in sources:
                    if isinstance(src, Path):
                        paths.append(src)
                    else:
                        try:
                            url = urlparse(src)
                            if url.scheme == 'file':
                                paths.append(Path(url.path))
                            else:
                                urls.append(src)
                        except ValueError:
                            msg.fail(f'Cannot parse source: {src}')

                yield from do_scan(paths, verbose=verbose, **kwargs)

                if urls:
                    yield from do_scan_with_syft(urls, verbose=verbose, **kwargs)

        except ExecutableNotFoundError as err:
            msg.fail(err)
            exit(2)

    scans = []
    for s in _do_scan():
        s.tag = tag
        s.branch = branch

        scans.append(process_scan(s))

    if scans:
        output_scans(scans, output_path, scan_format)


def output_scans(scans: list, path: t.Optional[Path], fmt: str = 'ts'):
    if path:
        with path.resolve().open('w') as fp:
            dump_scans(scans, fp, fmt)
    else:
        output = StringIO()
        dump_scans(scans, output, fmt)
        output.seek(0)
        print(output.read())



import click
import json
import typing as t

from tqdm import tqdm
from pathlib import Path
from concurrent import futures

from . import cli, load_scans_from_file
from .. import DependencyScan

from ..api import TrustSourceAPI


@cli.command('eval', help="Evaluates the scan result, may exit with non zero upon demand")
@cli.inout_default_options(_in=True, _out=True, _fmt=True)
@cli.api_default_options(project_name=False)
@click.option('--vulns-confidence', type=click.Choice(choices=['high', 'medium', 'low']), default='medium',
              help="Confidence level for matching CVEs to packages")
@click.option('--exit-with-failure-on-vulns', default=False, is_flag=True,
              help="Exit with a nonzero exit code (1) if vulnerabilities were found")
def eval_scan(path: Path,
              output_path: t.Optional[Path],
              scan_format: str,
              base_url: str,
              api_key: str,
              vulns_confidence: str,
              exit_with_failure_on_vulns: bool):

    if vulns_confidence == 'high':
        vulns_confidence = 1
    elif vulns_confidence == 'medium':
        vulns_confidence = 2
    else:
        vulns_confidence = 3

    ts_api = TrustSourceAPI(base_url, api_key)

    scans = load_scans_from_file(path, scan_format)
    vulns = {}

    for scan in scans:
        vulns.update(eval_vulns(scan, vulns_confidence, ts_api))

    if output_path:
        with output_path.resolve().open('w') as fp:
            json.dump(vulns, fp, indent=2) # noqa
    else:
        print(json.dumps(vulns, indent=2))

    if vulns and exit_with_failure_on_vulns:
        exit(1)


def eval_vulns(scan: DependencyScan, confidence: int, api: TrustSourceAPI) -> t.Dict[str, dict]:
    vulns = {}
    purls = [dep.purl.to_string() for dep in scan.iterdeps_once()]
    pbar = tqdm(desc="Checking vulnerabilities", total=len(purls))

    def check_completed(_task):
        result = _task.result()
        for res in result:
            comp_vulns = []
            for res_cves in res.get('cves', []):
                if (cves := res_cves['cves']) and cves and (res_cves.get('confidence', 0) < confidence):
                    comp_vulns.append(res_cves)

            if comp_vulns:
                vulns[res['key']] = comp_vulns

        pbar.update(len(result))

    tasks = []
    pool = futures.ThreadPoolExecutor()
    chunk_size = min(len(purls) // pool._max_workers, 40)

    for i in range(0, len(purls), chunk_size):
        task = pool.submit(lambda _purls: check_vulns(_purls, api), purls[i:i + chunk_size])
        task.add_done_callback(check_completed)
        tasks.append(task)

    futures.wait(tasks, return_when=futures.ALL_COMPLETED)
    return vulns




import json
import requests
import typing as t

from semantic_version import Version, NpmSpec

from pathlib import Path
from typing import Optional

from . import PackageManagerScanner, DependencyScan, Dependency, License


class NodeScanner(PackageManagerScanner):
    def __init__(self, enableMetadataRetrieval=False, **kwargs):
        super().__init__(**kwargs)

        self.enableMetadataRetrieval = enableMetadataRetrieval

        self.__path = None
        self.__abs_module_path = None

        self.__lookup = {}
        self.__failed_requests = 0
        self.__lockfile_content = {}
        self.__processed_deps = set()

    @staticmethod
    def name() -> str:
        return "Node"

    @staticmethod
    def executable() -> t.Optional[str]:
        return 'npm'

    @classmethod
    def options(cls) -> PackageManagerScanner.OptionsType:
        return super().options() | {
            'enableMetadataRetrieval': {
                'default': False,
                'is_flag': True,
                'help': 'Enable retrieving package metadata from the NPMJS online registry'
            }
        }

    def accepts(self, path: Path) -> bool:
        return path.is_dir() and (path / 'package.json').exists()

    def scan(self, path: Path) -> t.Optional[DependencyScan]:
        self.__path = path
        self.__abs_module_path = path.resolve().absolute()

        self._exec('install', cwd=self.__path)
        lock_file = self.__path / "package-lock.json"

        if not lock_file.exists():
            return

        with lock_file.open() as lockfile:
            self.__lockfile_content = json.load(lockfile)

        # extract root
        root = self.__lockfile_content['packages'].pop('', None)

        # first build dictionary that is indexable by name
        self.__lookup = self._dict_from_lock(self.__lockfile_content)

        # then traverse graph as described in lockfile
        pkgs = self.__lockfile_content['packages'].keys()

        if deps := [self._dep_from_lock(self.__lockfile_content, pkg_path) for pkg_path in pkgs]:
            module = ''
            moduleId = ''

            if root and (module := root.get('name')):
                moduleId = 'npm:' + module
                if version := root.get('version'):
                    moduleId += ':' + version

            return DependencyScan(module=module, moduleId=moduleId, dependencies=deps)
        else:
            return None

    @staticmethod
    def _dep_name_from_path(package_path: str) -> str:
        return package_path.split("node_modules/")[-1]

    @staticmethod
    def _dict_from_lock(lock: dict) -> dict:
        deps_dict = {}

        for dep_path, dep_dict in lock["packages"].items():
            name = NodeScanner._dep_name_from_path(dep_path)
            version = dep_dict["version"]

            deps_dict.setdefault(name, {})
            deps_dict[name][version] = dep_dict
            deps_dict[name][version]["_path"] = dep_path

        return deps_dict

    def _dep_from_lock(self, lock: dict, package_path: str = "") -> Dependency:
        pkg = lock["packages"][package_path]
        name = pkg.get("name")

        if not name:
            name = NodeScanner._dep_name_from_path(package_path)

        version = pkg["version"]

        dep = Dependency("npm:" + name, name, type='npm')
        dep.versions.append(version)

        if name + version not in self.__processed_deps:
            self.__processed_deps.add(name + version)

            dep.package_files.append(str(self.__abs_module_path / package_path))

            # msg.info(f"Getting metadata for {name} {version}...")
            if self.enableMetadataRetrieval and (meta := self._metadata_from_registry(name, version)):
                dep.licenses = meta.licenses
                dep.description = meta.description
                dep.homepageUrl = meta.homepageUrl
                dep.repoUrl = meta.repoUrl

            # msg.good(" Success!")
            # else:
            #   msg.fail(" Failed!")

            for dep_name, dep_version_range in pkg.get("dependencies", {}).items():
                dep_path = None

                # find appropriate dependency version
                for dep_version, dep_version_dict in self.__lookup[dep_name].items():
                    dep_path = dep_version_dict["_path"]

                    if dep_version_range == "latest":
                        break

                    try:
                        range_spec = NpmSpec(dep_version_range)

                    except ValueError:
                        range_spec = NpmSpec("")

                    if Version(dep_version) in range_spec:
                        break

                if dep_path:
                    dep.dependencies.append(
                        self._dep_from_lock(self.__lockfile_content, dep_path)
                    )

        return dep

    def _metadata_from_registry(self, name: str, version: str) -> Optional[Dependency]:
        template = "https://registry.npmjs.org/{}/{}"

        try:
            result = requests.get(template.format(name, version))

            if result.status_code != 200:
                raise requests.ConnectionError()

            meta = json.loads(result.text)

        except requests.ConnectionError:
            self.__failed_requests += 1
            return None

        dep = Dependency(key=f"npm:{name}", name=name, type='npm')

        dep.versions.append(meta.get("version", ""))

        dep.description = meta.get("description", "")
        dep.homepageUrl = meta.get("homepage", "")
        dep.repoUrl = meta.get("repository", {}).get("url", "")

        dep.licenses.append(License(meta.get("license")))

        return dep



import requests
import typing as t


class TrustSourceAPI:
    class Error(Exception):
        pass

    def __init__(self, base_url: str, api_key: str):
        self.__base_url = base_url
        self.__headers = {
            'Content-Type': 'application/json',
            'user-agent': f'ts-scan/1.1.0',
            'x-api-key': api_key,
            # 'X-APIKEY': api_key
        }

    def _post(self, path: str, data: dict, base_url=None) -> dict:
        if not base_url:
            base_url = self.__base_url
        resp = requests.post(f'{base_url}/{path}', json=data, headers=self.__headers)
        if resp:
            return resp.json()
        else:
            raise TrustSourceAPI.Error(resp.text)

    def upload_scan(self, data: dict) -> dict:
        return self._post('v2/core/scans', data)

    def find_cves(self, comps: t.List[str]) -> t.List[dict]:
        return self._post('v2/vulnerabilities/cveFindByKey', {
            'comps': [{'key': comp} for comp in comps]
        }).get('comps', [])

    def check_components(self, proj_name: str, mod_name: str, comps: t.List[dict]) -> dict:
        return self._post('v2/compliance/check/component', {
            'projectName': proj_name,
            'moduleName': mod_name,
            'components': comps
        })



import typing as t
from typing_extensions import TextIO

from copy import copy
from uuid import uuid4
from datetime import datetime
from pathlib import Path

from packageurl import PackageURL
from license_expression import Licensing, LicenseExpression, ExpressionError, combine_expressions
from spdx_tools.spdx.model import (Package, Document, CreationInfo, Actor, ActorType,
                                   ExternalPackageRef, ExternalPackageRefCategory, SpdxNone,
                                   Relationship, RelationshipType)
from spdx_tools.spdx.formats import FileFormat, file_name_to_format

from ..pm import Dependency, DependencyScan, License
from ..cli import msg


def import_scan(path: Path, fmt: str) -> t.Optional[DependencyScan]:
    _fmt = file_name_to_format(fmt.replace('-', '.'))

    if doc := _parse_file(path, _fmt):
        return DependencyScan(module=doc.creation_info.name,
                              moduleId=f'spdx:{doc.creation_info.name}',
                              dependencies=_create_deps(doc))

    else:
        raise ValueError('Cannot parse SPDX document')


def export_scan(scan: DependencyScan, output: t.TextIO, fmt: str):
    creation_info = CreationInfo(
        spdx_id='SPDXRef-DOCUMENT',
        spdx_version='SPDX-2.3',
        name=scan.module,
        document_namespace=f'https://spdx.org/spdxdocs/{scan.module}-{str(uuid4())}',
        creators=[Actor(ActorType.TOOL, 'ts-scan')],
        created=datetime.now()
    )

    packages, relationships = _create_pkgs(scan.dependencies)
    doc = Document(creation_info=creation_info,
                   packages=packages,
                   relationships=relationships)

    _fmt = file_name_to_format(fmt.replace('-', '.'))
    _write_stream(doc, output, _fmt)


def _create_pkgs(deps: t.Iterable[Dependency]) -> t.Tuple[t.List[Package], t.List[Relationship]]:
    visited: t.Dict[t.Tuple[str, t.Optional[str]], Package] = {}

    def _create_pkgs_impl(_deps: t.Iterable[Dependency]) -> t.Tuple[t.List[Package], t.List[Relationship]]:
        packages = []
        relations = []

        for dep in _deps:
            pkg_key = dep.key, dep.version
            pkg = visited.get(pkg_key)

            if not pkg:
                pkg = _create_pkg(dep, len(visited))
                visited[pkg_key] = pkg

            packages.append(pkg)

            pkgs, rels = _create_pkgs_impl(dep.dependencies)
            relations.extend(rels)

            for p in pkgs:
                relations.append(Relationship(pkg.spdx_id, RelationshipType.DEPENDS_ON, p.spdx_id))

        return packages, relations

    _, relationships = _create_pkgs_impl(deps)
    return list(visited.values()), relationships


def _create_pkg(dep: Dependency, ref_id: int) -> Package:
    pkg = Package(name=dep.name, spdx_id=f'SPDXRef-{ref_id}', download_location=SpdxNone())
    pkg.version = dep.version

    licensing = Licensing()
    lic_exprs: t.List[LicenseExpression] = []

    for lic in dep.licenses:
        try:
            lic_exprs.append(licensing.parse(lic.name, simple=True))
        except ExpressionError:
            msg.warn(f'Could not parse license expression "{lic.name}" of the dependency "{dep.key}"')

    if not lic_exprs:
        pkg.license_declared = SpdxNone()
    else:
        pkg.license_declared = combine_expressions(lic_exprs, relation='OR')

    pkg_ref = ExternalPackageRef(
        category=ExternalPackageRefCategory.PACKAGE_MANAGER,
        reference_type='purl',
        locator=dep.purl.to_string()
    )
    pkg.external_references.append(pkg_ref)

    return pkg


def _create_deps(doc: Document) -> t.List[Dependency]:
    deps = {}
    visited: t.Dict[str, Dependency] = {}

    for pkg in doc.packages:
        if dep := _create_dep(pkg):
            deps[pkg.spdx_id] = dep

    for rel in doc.relationships:
        if rel.relationship_type == RelationshipType.DEPENDS_ON and (type(rel.related_spdx_element_id) is str):
            src = deps.get(rel.spdx_element_id)
            if not src:
                src = visited.get(rel.spdx_element_id)

            if dst := deps.pop(rel.related_spdx_element_id, None):
                visited[rel.related_spdx_element_id] = dst
            else:
                dst = visited.get(rel.related_spdx_element_id)

            if src and dst:
                dst = copy(dst)
                dst.dependencies = []

                src.dependencies.append(dst)

    visited_deps: t.Dict[t.Tuple[str, t.Optional[str]], Dependency] = {
        (dep.key, dep.version): dep for dep in visited.values()
    }

    stack: t.List[Dependency] = []
    stack.extend(deps.values())

    while len(stack) > 0:
        cur = stack.pop()
        cur_deps = []

        for dep in cur.dependencies:
            if d := visited_deps.pop((dep.key, dep.version), None):
                cur_deps.append(d)
                stack.append(d)
            else:
                cur_deps.append(dep)

        cur.dependencies = cur_deps

    return list(deps.values())


def _create_dep(pkg: Package) -> t.Optional[Dependency]:
    dep = None
    pkg_mngr_info = None

    for ref in pkg.external_references:
        if ref.category == ExternalPackageRefCategory.PACKAGE_MANAGER:
            pkg_mngr_info = ref

    if pkg_mngr_info and pkg_mngr_info.reference_type == 'purl':
        try:
            purl = PackageURL.from_string(pkg_mngr_info.locator)
        except ValueError:
            purl = None

        if pkg.version and purl and (dep := Dependency.create_from_purl(purl)):
            dep.versions.append(pkg.version)

    if dep and pkg.license_declared:
        if lic_expr := pkg.license_declared:
            if isinstance(lic_expr, LicenseExpression):
                dep.licenses = [License(name=str(s)) for s in lic_expr.symbols]

    return dep


def _write_stream(document: Document, stream: TextIO, fmt: FileFormat, validate=False):
    from spdx_tools.spdx.writer.json import json_writer
    from spdx_tools.spdx.writer.tagvalue import tagvalue_writer
    from spdx_tools.spdx.writer.xml import xml_writer
    from spdx_tools.spdx.writer.yaml import yaml_writer

    if fmt == FileFormat.JSON:
        json_writer.write_document_to_stream(document, stream, validate)
    elif fmt == FileFormat.YAML:
        yaml_writer.write_document_to_stream(document, stream, validate)
    elif fmt == FileFormat.XML:
        xml_writer.write_document_to_stream(document, stream, validate)
    elif fmt == FileFormat.TAG_VALUE:
        tagvalue_writer.write_document_to_stream(document, stream, validate)


def _parse_file(path: Path, fmt: FileFormat) -> t.Optional[Document]:
    from spdx_tools.spdx.parser.json import json_parser
    from spdx_tools.spdx.parser.tagvalue import tagvalue_parser
    from spdx_tools.spdx.parser.xml import xml_parser
    from spdx_tools.spdx.parser.yaml import yaml_parser

    if fmt == FileFormat.TAG_VALUE:
        return tagvalue_parser.parse_from_file(str(path), 'utf-8')
    elif fmt == FileFormat.JSON:
        return json_parser.parse_from_file(str(path), 'utf-8')
    elif fmt == FileFormat.XML:
        return xml_parser.parse_from_file(str(path), 'utf-8')
    elif fmt == FileFormat.YAML:
        return yaml_parser.parse_from_file(str(path), 'utf-8')

    return None



import click
import typing as t

from pathlib import Path

from . import cli, scan_formats
from .scan import output_scans
from ..pm import load_scans


@cli.command('convert', help='Converts SBOM elements from one format to another')
@cli.inout_default_options(_in=True, _out=True, _fmt=True)
@click.option('-of', '--output-format', 'output_format',
              type=click.Choice(choices=scan_formats), default='ts', help='Output SBOM format')
def convert(path: Path,
            output_path: t.Optional[Path],
            scan_format: str,
            output_format: str):
    if scans := load_scans(path, scan_format):
        output_scans(scans, output_path, output_format)



import click
import typing as t

from pathlib import Path

from .pm import Scanner, Dependency, DependencyScan
from .cli import cli, msg


__version__ = '1.1.0'


def __get_pm_scanner_classes() -> t.List[t.Type[Scanner]]:
    from .pm.pypi import PypiScanner
    from .pm.maven import MavenScanner
    from .pm.node import NodeScanner
    from .pm.nuget import NugetScanner

    return [
        PypiScanner,
        MavenScanner,
        NodeScanner,
        NugetScanner
    ]


def scanner_options(f):
    from .pm.syft import SyftScanner

    scanner_classes = __get_pm_scanner_classes()
    scanner_classes.append(SyftScanner)

    for cls in scanner_classes:
        for opt, opt_params in cls.options().items():
            opt_prefix = cls.name().lower()
            f = click.option(f'--{opt_prefix}:{opt}', f'{opt_prefix}_{opt}', **opt_params)(f)

    return f


cli.scanner_options = scanner_options


def create_scanners(scanner_classes: t.Iterable[t.Type[Scanner]], **kwargs) -> [Scanner]:
    scanner_args = {cls.name().lower(): {} for cls in scanner_classes}
    other_args = {}

    for arg, val in kwargs.items():
        scanner_prefix_pos = arg.find('_')
        if scanner_prefix_pos >= 0:
            scanner_prefix = arg[:scanner_prefix_pos]
            scanner_arg = arg[scanner_prefix_pos + 1:]
            if _scanner_args := scanner_args.get(scanner_prefix):
                _scanner_args[scanner_arg] = val
        else:
            other_args[arg] = val

    return [cls(**other_args, **scanner_args[cls.name().lower()]) for cls in scanner_classes]


def do_scan(paths: t.List[Path], **kwargs) -> t.Iterable[DependencyScan]:
    """
    Excutes actual scan routines
    :param paths: List of paths to be scanned
    :return: An iterable over scan results
    """
    scanners = [s for s in create_scanners(__get_pm_scanner_classes(), **kwargs) if not s.ignore]

    for p in paths:
        p = p.resolve()

        msg.info(f'Running dependency scan.')

        scanned_at_least_once = False

        for scanner in scanners:
            if not scanner.accepts(p):
                continue

            scanned_at_least_once = True

            # with msg.loading(f'Scanning for {name} dependencies...'):
            msg.info(f'Found {scanner.name()} project. Scanning for dependencies...')

            if scan := _execute_scan(p, scanner):
                yield scan

            msg.good(f'{scanner.name()} scan is done!')

        if scanned_at_least_once:
            msg.good(f'Dependency scan completed.')
        else:
            msg.warn(f'No supported projects found.')


def do_scan_with_syft(sources: t.List[t.Union[Path, str]], **kwargs) -> t.Iterable[DependencyScan]:
    from .pm.syft import SyftScanner

    if scanners := create_scanners([SyftScanner], **kwargs):
        for src in sources:

            msg.info(f'Scanning for dependencies using Syft...')

            if scan := _execute_scan(src, scanners[0]):
                yield scan


def _execute_scan(src: t.Union[Path, str], scanner: Scanner) -> t.Optional[DependencyScan]:
    try:
        return scanner.scan(src)
    except Exception as err:
        msg.fail(f'An error occured while scanning {scanner.name()} dependencies...')

        if len(err.args) > 1:
            msg.fail(err.args[1])
        else:
            msg.fail(err)

        return None


def process_scan(scan: DependencyScan) -> DependencyScan:
    for dep in scan.iterdeps():
        dep.meta['purl'] = dep.purl.to_string()

    return scan
